<table id="main">
  <tr id="cluster_title_row" {{bind-attr style=Pcs.permissions_page}}>
    <td id="page_header" colspan="2" style="vertical-align:middle;">
      <table id="permissions_commands">
        <tr>
          <td>PERMISSIONS</td>
        </tr>
      </table>
    </td>
  </tr>
  <tr id="cluster_header_row" {{bind-attr style=Pcs.permissions_page}}>
    <td class="borderbottom" colspan="2"> </td>
  </tr>
  <tr id="cluster_list_row" {{bind-attr style=Pcs.permissions_page}}>
    <td id="cluster_list" class="node_list" style="width:auto;">
      <table cellpadding="0" cellspacing="0" style="width:100%;">
        <tr>
          <th> </th>
          <th> </th>
          <th>CLUSTER&nbsp;NAME</th>
          <th style="padding-right: 16px;"> </th>
        </tr>
        <% @clusters.each do |c| %>
          <tr
            onmouseover="if(permissions_current_cluster != '<%= h(c.name) %>') hover_over(this);"
            onmouseout="if(permissions_current_cluster != '<%= h(c.name) %>') hover_out(this);"
            onclick="permissions_show_cluster('<%= h(c.name) %>', this);"
          >
            <td class="node_list_sprite">
              <div class="check sprites"></div>
            </td>
            <td style="min-width:1em; padding-right:0.5em;">
              <span class="unsaved_changes" style="display:none;"
                title="There are unsaved changes in the form"
              >(*)</span>
            </td>
            <td nowrap class="node_name">
              <%= h(c.name) %>
            </td>
            <td>
              <div class="arrow sprites" style="display:none;"></div>
            </td>
          </tr>
        <% end %>
      </table>
    </td>
    <td id="node_info">
      <% @clusters.each { |c| %>
        <div id="permissions_cluster_<%= h(c.name) %>" style="display:none;">
        </div>
      <% } %>
    </td>
  </tr>
</table>
<%= erb :_dialogs %>
