from typing import NewType

AddRemoveContainerType = NewType("AddRemoveContainerType", str)
AddRemoveItemType = NewType("AddRemoveItemType", str)
BoothConfigUsedWhere = NewType("BoothConfigUsedWhere", str)
DefaultAddressSource = NewType("DefaultAddressSource", str)
FenceHistoryCommandType = NewType("FenceHistoryCommandType", str)
ForceCode = NewType("ForceCode", str)
MessageCode = NewType("MessageCode", str)
PcsCommand = NewType("PcsCommand", str)
ReasonType = NewType("ReasonType", str)
ServiceAction = NewType("ServiceAction", str)
SeverityLevel = NewType("SeverityLevel", str)
StonithRestartlessUpdateUnableToPerformReason = NewType(
    "StonithRestartlessUpdateUnableToPerformReason", str
)
